/*
 *  This file was modified for VietPad from the original supplied by Unicode, Inc.,
 *  adding a normalize method accepting a char parameter and returning a char
 *  and static compose and decompose methods matching com.ibm.icu.text.Normalizer.
 */

using System;
using System.Text;

namespace org.unicode
{
	
	/// <summary> Implements Unicode Normalization Forms C, D, KC, KD.<br>
	/// See UTR#15 for details.<br>
	/// Copyright (c) 1998-1999 Unicode, Inc. All Rights Reserved.<br>
	/// The Unicode Consortium makes no expressed or implied warranty of any
	/// kind, and assumes no liability for errors or omissions.
	/// No liability is assumed for incidental and consequential damages
	/// in connection with or arising out of the use of the information here.
	/// </summary>
	/// <author> Mark Davis
	/// 
	/// </author>
	
	public class Normalizer
	{
		internal const string copyright = "Copyright (c) 1998-1999 Unicode, Inc.";
		
		/// <summary> Create a normalizer for a given form.
		/// </summary>
		public Normalizer(sbyte form)
		{
			this.form = form;
			if (data == null)
				data = NormalizerBuilder.build();
			// load 1st time
		}
		
		/// <summary> Masks for the form selector
		/// </summary>
		internal const sbyte COMPATIBILITY_MASK = 1;
		internal const sbyte COMPOSITION_MASK = 2;
		
		/// <summary> Normalization Form Selector
		/// </summary>
		public const sbyte D = 0;
		public static sbyte C = COMPOSITION_MASK;
		public static sbyte KD = COMPATIBILITY_MASK;
		public static sbyte KC = (sbyte) (COMPATIBILITY_MASK + COMPOSITION_MASK);
		
		/// <summary> Normalizes text according to the chosen form, 
		/// replacing contents of the target buffer.
		/// </summary>
		/// <param name="source">     the original text, unnormalized
		/// </param>
		/// <param name="target">     the resulting normalized text
		/// 
		/// </param>
		public StringBuilder normalize(string source, StringBuilder target)
		{
			
			// First decompose the source into target,
			// then compose if the form requires.
			
			if (source.Length != 0)
			{
				internalDecompose(source, target);
				if ((form & COMPOSITION_MASK) != 0)
				{
					internalCompose(target);
				}
			}
			return target;
		}
		
		/// <summary> Normalizes text according to the chosen form
		/// </summary>
		/// <param name="source">     the original text, unnormalized
		/// </param>
		/// <returns> target      the resulting normalized text
		/// 
		/// </returns>
		public string normalize(string source)
		{
			return normalize(source, new StringBuilder()).ToString();
		}
		
		/// <summary> Normalizes text according to the chosen form
		/// </summary>
		/// <param name="source">     the original char, unnormalized
		/// </param>
		/// <returns> target      the resulting normalized char
		/// 
		/// </returns>
		public char normalize(char source)
		{
			return normalize(source.ToString())[0];
		}

		/**
		 *  Composes a string. The string will be composed to according to the
		 *  specified mode.
		 *
		 *@param  str     The string to compose.
		 *@param  compat  If true the string will be composed accoding to NFKC rules
		 *      and if false will be composed according to NFC rules.
		 *@return         string The composed string
		 *@draft          ICU 2.6
		 */
		public static string compose(string str, bool compat) 
		{

			if (normalizer == null) 
			{
				normalizer = new Normalizer(compat ? KC : C);
			} 
			else 
			{
				normalizer.form = compat ? KC : C;
			}
			return normalizer.normalize(str);
		}


		/**
		 *  Decomposes a string. The string will be decomposed to according to the
		 *  specified mode.
		 *
		 *@param  str     The string to decompose.
		 *@param  compat  If true the string will be decomposed accoding to NFKD rules
		 *      and if false will be decomposed according to NFD rules.
		 *@return         string The decomposed string
		 *@draft          ICU 2.6
		 */
		public static string decompose(string str, bool compat) 
		{

			if (normalizer == null) 
			{
				normalizer = new Normalizer(compat ? KD : D);
			} 
			else 
			{
				normalizer.form = compat ? KD : D;
			}
			return normalizer.normalize(str);
		}
		
		// ======================================
		//                  PRIVATES
		// ======================================
		
		static private Normalizer normalizer;

		/// <summary> The current form.
		/// </summary>
		private sbyte form;
		
		/// <summary> Decomposes text, either canonical or compatibility,
		/// replacing contents of the target buffer.
		/// </summary>
		/// <param name="form">       the normalization form. If COMPATIBILITY_MASK
		/// bit is on in this byte, then selects the recursive 
		/// compatibility decomposition, otherwise selects
		/// the recursive canonical decomposition.
		/// </param>
		/// <param name="source">     the original text, unnormalized
		/// </param>
		/// <param name="target">     the resulting normalized text
		/// 
		/// </param>
		private void  internalDecompose(string source, StringBuilder target)
		{
			StringBuilder buffer = new StringBuilder();
			bool canonical = (form & COMPATIBILITY_MASK) == 0;
			for (int i = 0; i < source.Length; ++i)
			{
				buffer.Length = 0;
				data.getRecursiveDecomposition(canonical, source[i], buffer);
				
				// add all of the characters in the decomposition.
				// (may be just the original character, if there was
				// no decomposition mapping)
				
				for (int j = 0; j < buffer.Length; ++j)
				{
					char ch = buffer[j];
					int chClass = data.getCanonicalClass(ch);
					int k = target.Length; // insertion point
					if (chClass != 0)
					{
						
						// bubble-sort combining marks as necessary
						
						 for (; k > 0; --k)
						{
							if (data.getCanonicalClass(target[k - 1]) <= chClass)
								break;
						}
					}
					target.Insert(k, ch);
				}
			}
		}
		
		/// <summary> Composes text in place. Target must already
		/// have been decomposed.
		/// </summary>
		/// <param name="target">     input: decomposed text.
		/// output: the resulting normalized text.
		/// 
		/// </param>
		private void  internalCompose(StringBuilder target)
		{
			
			int starterPos = 0, compPos = 1;
			char starterCh = target[0];
			int lastClass = data.getCanonicalClass(starterCh);
			if (lastClass != 0)
				lastClass = 256;
			// fix for irregular combining sequence
			
			// Loop on the decomposed characters, combining where possible
			
			for (int decompPos = 1; decompPos < target.Length; ++decompPos)
			{
				char ch = target[decompPos];
				int chClass = data.getCanonicalClass(ch);
				char composite = data.getPairwiseComposition(starterCh, ch);
				if (composite != org.unicode.NormalizerData.NOT_COMPOSITE && (lastClass < chClass || lastClass == 0))
				{
					target[starterPos] = composite;
					starterCh = composite;
				}
				else
				{
					if (chClass == 0)
					{
						starterPos = compPos;
						starterCh = ch;
					}
					lastClass = chClass;
					target[compPos++] = ch;
				}
			}
			target.Length = compPos;
		}
		
		/// <summary> Contains normalization data from the Unicode Character Database.
		/// use false for the minimal set, true for the real set.  
		/// </summary>
		private static NormalizerData data = null;
		
		/// <summary> Just accessible for testing.
		/// </summary>
		internal bool getExcluded(char ch)
		{
			return data.getExcluded(ch);
		}
		
		/// <summary> Just accessible for testing.
		/// </summary>
		internal string getRawDecompositionMapping(char ch)
		{
			return data.getRawDecompositionMapping(ch);
		}
	}
}
